from abc import ABC, abstractmethod


class ParameterStore(ABC):
    @abstractmethod
    def get_param(self, param: str) -> str:
        raise NotImplementedError

    @abstractmethod
    def set_param(
        self, param: str, value: str, overwrite: bool = False, secure: bool = False
    ) -> None:
        raise NotImplementedError



from abc import ABC, abstractmethod
from typing import Dict

from pyspark.sql import SparkSession


class SparkSessions(ABC):
    @abstractmethod
    def enable_hive_support(self) -> bool:
        raise NotImplementedError

    @abstractmethod
    def spark_options(self) -> Dict[str, str]:
        raise NotImplementedError

    @staticmethod
    def default_configuration() -> Dict[str, str]:
        return {
            "spark.serializer": "org.apache.spark.serializer.KryoSerializer",
            "spark.sql.avro.compression.codec": "snappy",
            "spark.sql.parquet.compression.codec": "snappy",
            "spark.sql.sources.partitionColumnTypeInference.enabled": "false",
            "spark.sql.sources.partitionOverwriteMode": "dynamic",
        }

    def spark_session_builder(self) -> SparkSession.Builder:
        builder = SparkSession.Builder()

        joined_config: Dict[str, str] = self.spark_options().copy()
        joined_config.update(self.default_configuration())
        for key, value in joined_config.items():
            builder.config(key, value)

        return builder

    def spark(self) -> SparkSession:
        builder = self.spark_session_builder()

        if self.enable_hive_support():
            builder.enableHiveSupport()

        return builder.getOrCreate()



from abc import ABC

from pyspark.sql import SparkSession

from .data_link import DataLink


class PathBasedDataLink(DataLink, ABC):
    def __init__(self, environment: str, session: SparkSession, path: str):
        super().__init__(environment, session)
        self.path = path



from pyspark.sql import SparkSession, DataFrame

from .data_link import DataLink


class SnowflakeDataLink(DataLink):
    def __init__(
        self,
        environment: str,
        session: SparkSession,
        account: str,
        username: str,
        password: str,
        database: str,
        schema: str,
        table: str,
        warehouse: str,
        role: str,
        parallelism: int = 1,
        save_mode: str = "error",
    ):
        super().__init__(environment, session)

        self.save_mode = save_mode

        self.sfOptions = {
            "sfURL": f"{account}.snowflakecomputing.com",
            "sfUser": username,
            "sfPassword": password,
            "sfDatabase": database,
            "sfSchema": schema,
            "sfWarehouse": warehouse,
            "dbtable": table,
            "parallelism": parallelism,
            "role": role,
        }

        self.connection_format = "net.snowflake.spark.snowflake"

    def read(self) -> DataFrame:
        # Partition parameters are only applicable for read operation
        # for now, order is important as some values can be overwritten
        return self.spark.read.format(self.connection_format).options(**self.sfOptions).load()

    def write(self, frame: DataFrame) -> None:
        (
            frame.write.format(self.connection_format)
            .mode(self.save_mode)
            .options(**self.sfOptions)
            .save()
        )



from typing import Mapping, Optional, List

from pyspark.sql import SparkSession, DataFrame

from pyhouse.datalake.data_link import DataLink


class KafkaDataLink(DataLink):
    def __init__(
        self,
        environment: str,
        session: SparkSession,
        topic: str,
        bootstrap_servers: List[str],
        options: Optional[Mapping[str, str]] = None,
    ):
        super().__init__(environment, session)

        self.topic = topic
        self.bootstrap_servers = bootstrap_servers
        self.options = {} if options is None else options

    def read(self) -> DataFrame:
        return (
            self.spark
            .read
            .format("kafka")
            .option("subscribe", self.topic)
            .option("kafka.bootstrap.servers", ",".join(self.bootstrap_servers))
            .options(**self.options)
            .load()
        )

    def write(self, frame: DataFrame) -> None:
        print(self.topic)
        (
            frame
            .write
            .format("kafka")
            .option("topic", self.topic)
            .option("kafka.bootstrap.servers", ",".join(self.bootstrap_servers))
            .options(**self.options)
            .save()
        )



from typing import Mapping, Sequence, Optional

from pyspark.sql import SparkSession, DataFrame
from pyspark.sql.types import StructType

from .path_based_data_link import PathBasedDataLink


class FileSystemDataLink(PathBasedDataLink):
    def __init__(
        self,
        environment: str,
        session: SparkSession,
        path: str,
        format: str,
        savemode: str,
        partitioned_by: Optional[Sequence[str]] = None,
        options: Optional[Mapping[str, str]] = None,
        schema: Optional[StructType] = None,
    ):
        super().__init__(environment, session, path)
        self.spark_format = format
        self.savemode = savemode
        self.partitioned_by = partitioned_by
        self.options = {} if options is None else options
        self.schema = schema

    def read(self) -> DataFrame:
        if self.schema is not None:
            return (
                self.spark.read.format(self.spark_format)
                .options(**self.options)
                .schema(self.schema)
                .load(self.path)
            )
        else:
            return (
                self.spark.read.format(self.spark_format)
                .options(**self.options)
                .load(self.path)
            )

    def write(self, frame: DataFrame) -> None:
        if self.partitioned_by is not None:
            (
                frame.write.format(self.spark_format)
                .partitionBy(self.partitioned_by)
                .options(**self.options)
                .mode(self.savemode)
                .save(self.path)
            )
        else:
            (
                frame.write.format(self.spark_format)
                .options(**self.options)
                .mode(self.savemode)
                .save(self.path)
            )



from abc import ABC, abstractmethod

from pyspark.sql import SparkSession, DataFrame


class DataLink(ABC):
    def __init__(self, environment: str, session: SparkSession):
        self.spark = session
        self.environment = environment

    @abstractmethod
    def read(self) -> DataFrame:
        raise NotImplementedError

    @abstractmethod
    def write(self, frame: DataFrame) -> None:
        raise NotImplementedError



from typing import Callable

from pyspark.sql import DataFrame

from .data_link import DataLink


def sink(self: DataFrame, dl: DataLink) -> None:
    """Write a DataFrame to a DataLink.
    Spark SQL allows ETL jobs to be written using chained transformations, like
    `spark.read.parquet().transform(func1).transform(func2).write(destination)`
    where `destination` is a string. The DataLinks introduced in LightHouse do
    not support this intuitive paradigm, unless DataFrames are monkey-patched.
    """
    dl.write(frame=self)


def transform(self: DataFrame, fun: Callable[[DataFrame], DataFrame]) -> DataFrame:
    """Monkey-patch pyspark DataFrames so they have a `transform`
    method as in (Scala) Spark."""
    return fun(self)


DataFrame.sink = sink
DataFrame.transform = transform



from .data_link import DataLink
from pyspark.sql import DataFrame, SparkSession


class InMemoryDataLink(DataLink):
    def __init__(self, environment: str, session: SparkSession, df: DataFrame):
        super(InMemoryDataLink, self).__init__(environment, session)
        self.df = df

    def read(self) -> DataFrame:
        return self.df

    def write(self, frame: DataFrame) -> None:
        self.df = frame



import logging
import boto3

from .parameter_store import ParameterStore

logger = logging.getLogger(__name__)


class AwsParameterStore(ParameterStore):
    """
    Retrieve or store values from/into the AWS Parameter Store.
    """

    def __init__(self, region_name: str = "eu-west-1"):
        self.ssm = boto3.client("ssm", region_name=region_name, verify=None)

    def get_param(self, name: str) -> str:
        # aws will randomly drop ssm requests
        try:
            value = self.ssm.get_parameter(Name=name, WithDecryption=True)["Parameter"]["Value"]
            return str(value)
        except Exception as e:
            logger.warning(f"AwsParameterStore::get_param:{name}: Caught exception {e}")
            raise e

    def set_param(
        self, param: str, value: str, overwrite: bool = False, secure: bool = False
    ) -> None:
        param_type = "SecureString" if secure else "String"
        try:
            # aws will randomly drop ssm requests
            self.ssm.put_parameter(Name=param, Value=value, Type=param_type, Overwrite=overwrite)
        except Exception as e:
            logger.warning("AwsParameterStore::set_param: Caught exception")
            raise e



import pyspark.sql.functions as sfun
from pyspark.sql import SparkSession, DataFrame

from .data_link import DataLink


class JdbcDataLink(DataLink):
    def __init__(
        self,
        environment: str,
        session: SparkSession,
        url: str,
        username: str,
        password: str,
        driver: str,
        table: str,
        save_mode: str = "error",
        number_of_partitions: int = 1,
        partition_column: str = "",
    ):
        super().__init__(environment, session)

        self.number_of_partitions = number_of_partitions
        self.partition_column = partition_column
        self.save_mode = save_mode
        self.connection_properties = {
            "url": url,
            "user": username,
            "password": password,
            "driver": driver,
            "dbtable": table,
        }

    def read(self) -> DataFrame:
        # Partition parameters are only applicable for read operation
        # for now, order is important as some values can be overwritten
        reader = self.spark.read.format("jdbc").options(**self.connection_properties)

        if self.number_of_partitions == 1:
            return reader.load()

        else:
            # We need a partition column
            assert self.partition_column != ""
            # Retrieve lower and upper bound first to
            # determine the degree of parallelism
            lower_bound, upper_bound = (
                reader.load()
                .select(
                    sfun.min(sfun.col(self.partition_column)).alias("mmin"),
                    sfun.max(sfun.col(self.partition_column)).alias("mmax"),
                )
                .collect()[0]
            )

            return (
                reader.option("partitionColumn", self.partition_column)
                .option("numPartitions", str(self.number_of_partitions))
                .option("lowerBound", str(lower_bound))
                .option("upperBound", str(upper_bound))
                .load()
            )

    def write(self, frame: DataFrame) -> None:
        (
            frame.write.format("jdbc")
            .mode(self.save_mode)
            .options(**self.connection_properties)
            .save()
        )



from configparser import ConfigParser

from .parameter_store import ParameterStore


class FileParameterStore(ParameterStore):
    def __init__(self, filename: str):
        self.filename = filename
        self.config = ConfigParser()

    def get_param(self, param: str) -> str:
        self.config.read(self.filename)
        return self.config["config"][param]

    def set_param(
        self, param: str, value: str, overwrite: bool = False, secure: bool = False
    ) -> None:
        self.config.read(self.filename)
        if overwrite:
            self.config["config"][param] = value
        else:
            self.config["config"].setdefault("param", value)
        with open(self.filename, "w") as out:
            self.config.write(out)



from setuptools import setup, find_packages
from pipenv.project import Project
from pipenv.utils import convert_deps_to_pip

# read the contents of your README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

pfile = Project(chdir=False).parsed_pipfile
requirements = convert_deps_to_pip(pfile['packages'], r=False)
dev_requirements = convert_deps_to_pip(pfile['dev-packages'], r=False)

setup(
    name='pyhouse',
    version='0.0.10',
    description='Python Lighthouse',
    author='Dataminded',
    license="Apache-2.0",
    author_email='dev@dataminded.be',
    url='https://www.dataminded.be',
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=find_packages(),
    install_requires=requirements,
    test_requires=dev_requirements,
    package_data={
        'pyhouse': ['py.typed'],
    },
    zip_safe=False,
)



from typing import Mapping, Sequence, Optional

from pyspark.sql import SparkSession, DataFrame

from .path_based_data_link import PathBasedDataLink


class HiveDataLink(PathBasedDataLink):
    def __init__(
        self,
        environment: str,
        session: SparkSession,
        path: str,
        database: str,
        table: str,
        storage_format: str = "parquet",
        save_mode: str = "overwrite",
        partitioned_by: Optional[Sequence[str]] = None,
        overwrite_behavior: str = "full_overwrite",
        options: Optional[Mapping[str, str]] = None,
    ):
        super().__init__(environment, session, path)
        self.database = database
        self.table = table
        self.format = storage_format
        self.save_mode = save_mode
        self.partitioned_by = partitioned_by
        self.overwrite_behavior = overwrite_behavior
        self.options = {} if options is None else options

    def read(self) -> DataFrame:
        self.spark.catalog.setCurrentDatabase(self.database)
        return self.spark.read.table(self.table)

    def write(self, frame: DataFrame) -> None:
        self.spark.catalog.setCurrentDatabase(self.database)
        # TODO: implement the cases for the overwrite_behavior flag
        # right now, the simplest approach "overwrite all" is used.
        (
            frame.write.saveAsTable(
                name=self.table,
                format=self.format,
                mode=self.save_mode,
                partitionBy=self.partitioned_by,
                path=self.path,
                **self.options,
            )
        )
